/*
 * SealedSphereSim.java
 */

package EDU.gatech.cc.is.abstractrobot;

import java.awt.*;
import java.util.Enumeration;
import EDU.gatech.cc.is.util.*;
import EDU.gatech.cc.is.simulation.*;
import EDU.gatech.cc.is.communication.*;


/**
 * Implements SealedSphere for simulation.
 * You should see the specifications in SealedSphere for details on 
 * how to use these methods.
 * <P>
 * <A HREF="../COPYRIGHT.html">Copyright</A>
 * (c)1997, 1998 Tucker Balch
 *
 * @see SealedSphere
 * @author Tucker Balch
 * @version $Revision: 1.6 $
 */

public class SealedSphereSim extends Simple 
	implements SealedSphere, SimulatedObject
	{
        private CircularBuffer  trail;  // robot's trail
	private Vec2		position;	// location of the robot
	private Vec2		steer;		// robot heading
	private	double		speed;		// current speed
	private	Color		foreground, 	// colors for drawing
				background;
	private long		time;		// elapsed time
	private double		timed;		// double version of time
	private double		left, 		// used for drawing purposes
				right, 
				top, 
				bottom;
	private	SimulatedObject[] all_objects 	// to keep track of other
		= new SimulatedObject[0];	// objects in the simulation
	private	int		visionclass;	// how other robots see us
	
	public	static final boolean DEBUG = false;// set true for debug
						// messages

	/**
	 * Instantiate a <B>SealedSphereSim</B> object.  Be sure
	 * to also call init with proper values.
	 * @see SealedSphereSim#init
	 */
        public SealedSphereSim()
		{
		if (DEBUG) System.out.println("SealedSphereSim: instantiated.");
		// defaults
		position = new Vec2(0,0);
		steer = new Vec2(1,0);
		foreground = Color.black;
		background = Color.black;
		top = 10;
		bottom = -10;
		left = -10;
		right = 10;
		}	


        /**
         * Initialize a <B>SealedSphereSim</B> object.
	 * @param	yp	ignored.
	 * @param	tp	ignored.
	 * @param	ignore	ignored.
	 * @param	f	color 1 for the robot.
	 * @param	b	color 2 for the robot.
	 * @param	v	vision class of the robot
	 * @param	i	unique simulation id 
	 *			(NOT the same as player number!).
         */
	public void init(double xp, double yp, double tp, double ignore,
		Color f, Color b, int v, int i, long s)
		{
                trail = new CircularBuffer(100);
		if (DEBUG) System.out.println("SealedSphereSim: initialized"
			+" at "+xp+","+yp);

		/*--- set global parameters ---*/
		setID(i);
		position.setx(xp);
		position.sety(yp);
		foreground = f;
		background = b;
		time = 0;
		timed = 0;
		visionclass = v;
		setID(i);
		steer.sett(0);
		}


	/**
	 * Take a simulated step;
	 */
	public void takeStep(long time_increment, SimulatedObject[] all_objs)
		{
		if (DEBUG) System.out.println("SealedSphereSim.TakeStep()");

		/*--- keep pointer to the other objects ---*/
		all_objects = all_objs;

		/*--- update the time ---*/
		time += time_increment;
		double time_incd = ((double)time_increment)/1000;
		timed += time_incd;

		/*--- update the steering ---*/
		double sturn = Units.BestTurnRad(steer.t, desired_heading);
		if (Math.abs(sturn) > (MAX_STEER*time_incd))
			{
			if (sturn<0)
				sturn = -MAX_STEER*time_incd;
			else sturn = MAX_STEER*time_incd;
			}
		steer.sett(steer.t + sturn);

		/*--- compute velocity ---*/
		Vec2 velocity = new Vec2(steer.x, steer.y);
		velocity.setr(base_speed * desired_speed);

		/*--- compute a movement step ---*/
		Vec2 mvstep = new Vec2(velocity.x, velocity.y);
		mvstep.setr(mvstep.r * time_incd);

		/*--- test the new position to see if on top of obstacle ---*/
		Vec2 pp = new Vec2(position.x, position.y);
		boolean moveok = true;
		pp.add(mvstep);
		for (int i=0; i<all_objects.length; i++)
			{
			if (all_objects[i].isObstacle() && 
				(all_objects[i].getID() != unique_id))
				{
				Vec2 tmp = all_objects[i].getClosestPoint(pp);
				if (tmp.r < RADIUS)
					{
					moveok = false;
					break;
					}
				}
			}
		if (moveok) position.add(mvstep);

		/*--- test the new position to see if on top of pushable ---*/
		for (int i=0; i<all_objects.length; i++)
			{
			if (all_objects[i].isPushable() && 
				(all_objects[i].getID() != unique_id))
				{
				Vec2 tmp = all_objects[i].getClosestPoint(pp);
				if (tmp.r < RADIUS)
					{
					tmp.setr(RADIUS - tmp.r);
					all_objects[i].push(tmp, velocity);
					}
				}
			}
		}

	public boolean isObstacle()
		{
		return(true);
		}
	
	public boolean isPushable()
		{
		return(false);
		}
	
	public boolean isPickupable()
		{
		return(false);
		}
	
	public Vec2 getClosestPoint(Vec2 from)
		{
		Vec2 tmp = new Vec2(position.x, position.y);
		tmp.sub(from);
		if (tmp.r < RADIUS)
			tmp.setr(0);
		else
			tmp.setr(tmp.r-RADIUS);
		return(tmp);
		}

	public Vec2 getCenter(Vec2 from)
		{
		Vec2 tmp = new Vec2(position.x, position.y);
		tmp.sub(from);
		return(tmp);
		}

	public void push(Vec2 d, Vec2 v)
		{
		// sorry no pushee robots!
		}

	public void pickUp(SimulatedObject o)
		{
		// sorry no pickupee robots!
		}

	public void putDown(Vec2 p)
		{
		// sorry no put downee robots!
		}

	public void setVisionClass(int v)
		{
		visionclass = v;
		}

	public int getVisionClass()
		{
		return(visionclass);
		}

        /**
         * Draw the robot in a specific spot.
         */
        public void draw(Vec2 pos, Graphics g, int w, int h,
                double t, double b, double l, double r)
                {
                Vec2 old_pos = position;
                position = pos;
                draw(g,w,h,t,b,l,r);
                position = old_pos;
                }

        private Point last = new Point(0,0);
        /**
         * Draw the robot's Trail.
         */
        public void drawTrail(Graphics g, int w, int h,
                double t, double b, double l, double r)
                {
                top =t; bottom =b; left =l; right =r;
                if (DEBUG) System.out.println("draw "+
                        w + " " +
                        h + " " +
                        t + " " +
                        b + " " +
                        l + " " +
                        r + " ");
                double meterspp = (r - l) / (double)w;
                int xpix = (int)((position.x - l) / meterspp);
                int ypix = (int)((double)h - ((position.y - b) / meterspp));

                /*--- record the point ---*/
                Point p = new Point(xpix,ypix);
                if ((last.x!=xpix)||(last.y!=ypix))
                        trail.put(p);
                last = p;

                /*--- get the list of all points ---*/
                Enumeration point_list = trail.elements();

                /*--- draw the trail ---*/
                g.setColor(background);
                Point from = (Point)point_list.nextElement();
                while (point_list.hasMoreElements())
                        {
                        Point next = (Point)point_list.nextElement();
                        g.drawLine(from.x,from.y,next.x,next.y);
                        from = next;
                        }
                }


	private String display_string = "blank";
	/**
         * Set the String that is printed on the robot's display.
         * For simulated robots, this appears printed below the agent
         * when view "Robot State" is selected.
         * @param s String, the text to display.
         */
        public void setDisplayString(String s)
		{
		display_string = s;
		}


        /**
         * Draw the robot's state.
         */
        public void drawState(Graphics g, int w, int h,
                double t, double b, double l, double r)
                {
                top =t; bottom =b; left =l; right =r;
                if (DEBUG) System.out.println("draw "+
                        w + " " +
                        h + " " +
                        t + " " +
                        b + " " +
                        l + " " +
                        r + " ");
                double meterspp = (r - l) / (double)w;
                int radius = (int)(RADIUS / meterspp);
                int xpix = (int)((position.x - l) / meterspp);
                int ypix = (int)((double)h - ((position.y - b) / meterspp));

                /*--- draw State ---*/
                g.setColor(background);
                g.drawString(display_string,xpix+radius+3,ypix-radius);
                }


	/**
         * Set the length of the trail (in movement steps).
         * @param l int, the length of the trail.
         */
        public void setTrailLength(int l)
		{
		trail = new CircularBuffer(l);
		}


        /**
         * Clear the trail.
         */
        public void clearTrail()
		{
		trail.clear();
		}
        /**
         * Draw the robot's ID.  Note, since Pebbles robots
         * do not have a kin sensor, we just draw the unique_id.
         */
        public void drawID(Graphics g, int w, int h,
                double t, double b, double l, double r)
                {
                top =t; bottom =b; left =l; right =r;
                if (DEBUG) System.out.println("draw "+
                        w + " " +
                        h + " " +
                        t + " " +
                        b + " " +
                        b + " " +
                        l + " " +
                        r + " ");
                double meterspp = (r - l) / (double)w;
                int radius = (int)(RADIUS / meterspp);
                int xpix = (int)((position.x - l) / meterspp);
                int ypix = (int)((double)h - ((position.y - b) / meterspp));

                /*--- draw ID ---*/
                g.setColor(background);
                g.drawString(String.valueOf(getPlayerNumber(0))
                        ,xpix-radius,ypix-radius);
                }



	/**
	 * Draw the robot.
	 */
	public void draw(Graphics g, int w, int h,
		double t, double b, double l, double r)
		{
		top =t; bottom =b; left =l; right =r;
		if (DEBUG) System.out.println("draw "+
			w + " " +
			h + " " +
			t + " " +
			b + " " +
			l + " " +
			r + " ");
		double meterspp = (r - l) / (double)w;
		if (DEBUG) System.out.println("meterspp "+meterspp);
		int radius = (int)(RADIUS / meterspp);
		int steerd = (int)Units.RadToDeg(steer.t);
		int xpix = (int)((position.x - l) / meterspp);
		int ypix = (int)((double)h - ((position.y - b) / meterspp));
		if (DEBUG) System.out.println("robot at"+
			" at "+xpix+","+ypix);

		/*--- draw the main body ---*/
		g.setColor(foreground);
		g.fillOval(xpix - radius, ypix - radius,
				radius + radius, radius + radius);

		/*--- draw steer      ---*/
		g.setColor(background);
		int dirx = xpix + (int)((double)radius * Math.cos(steer.t));
		int diry = ypix + -(int)((double)radius * Math.sin(steer.t));
		g.drawLine(xpix, ypix, dirx, diry);
		}


	/**
	 * Clean up.
	 */
	public void quit()
		{
		}

	/**
	 * Gets time elapsed since the robot was instantiated. 
	 * Since this is simulation, it may not match real elapsed time.
	 */
	public long getTime()
		{
		return(time);
		}


	private	long	last_Opponentst = 0;
	private	Vec2	last_Opponents[];
	private	int	num_Opponents;
	/**
	Get an array of Vec2s that point egocentrically from the
	center of the robot to the Opponents currently sensed by the 
	robot
	@param timestamp only get new information if timestamp > than last call
			or timestamp == -1 .
	@return the sensed Opponents.
	*/
	public Vec2[] getOpponents(long timestamp)
		{
		if((timestamp > last_Opponentst)||(timestamp == -1))
			{
			if (timestamp != -1) last_Opponentst = timestamp;
			Vec2 tmp_objs[] = new Vec2[all_objects.length];
			num_Opponents = 0;
			int our_v_class = getVisionClass();
			/*--- check all objects ---*/
			for(int i = 0; i<all_objects.length; i++)
				{
				/*--- check if it's an opponent and not obstcle-*/
				int vc = all_objects[i].getVisionClass();
				if (our_v_class!=vc && vc!=0)
					{
					Vec2 tmp = all_objects[i].getClosestPoint(
							position);
					if (tmp.r<kin_rangeM)
					tmp_objs[num_Opponents++] = tmp;
					}
				}
			last_Opponents = new Vec2[num_Opponents];
			for(int i = 0; i<num_Opponents; i++)
			last_Opponents[i] = new Vec2(tmp_objs[i].x,
				tmp_objs[i].y);
			}
		Vec2[] retval = new Vec2[num_Opponents];
		for(int i = 0; i<num_Opponents; i++)
			retval[i] = new Vec2(last_Opponents[i].x,
				last_Opponents[i].y);
		return(retval);
		}


	private	long	last_Teammatest = 0;
	private	Vec2	last_Teammates[];
	private	int	vis_Teammates=0;
	private	int	num_Teammates=1; // include self
	/**
	 * Get an array of Vec2s that point egocentrically from the
	 * center of the robot to the teammates currently sensed by the 
	 * robot
	 * @param timestamp only get new information 
	 *		if timestamp > than last call or timestamp == -1 .
	 * @return the sensed teammates.
	 */
	public Vec2[] getTeammates(long timestamp)
		{
		if((timestamp > last_Teammatest)||(timestamp == -1))
			{
			if (timestamp != -1) last_Teammatest = timestamp;
			Vec2 tmp_objs[] = new Vec2[all_objects.length];
			vis_Teammates = 0;
			num_Teammates = 1; // include self
			int our_v_class = getVisionClass();
			/*--- check all objects ---*/
			for(int i = 0; i<all_objects.length; i++)
				{
				/*--- check if it's a teammate and not self ---*/
				if (all_objects[i].getVisionClass()==our_v_class &&
					(all_objects[i].getID() != unique_id))
					{
					Vec2 tmp = all_objects[i].getClosestPoint(
							position);
					num_Teammates++;

					// check if in range
					if (tmp.r<kin_rangeM)
					tmp_objs[vis_Teammates++] = tmp;
					}
				}
			last_Teammates = new Vec2[vis_Teammates];
			for(int i = 0; i<vis_Teammates; i++)
			last_Teammates[i] = new Vec2(tmp_objs[i].x,
				tmp_objs[i].y);
			}
		Vec2[] retval = new Vec2[vis_Teammates];
		for(int i = 0; i<vis_Teammates; i++)
			retval[i] = new Vec2(last_Teammates[i].x,
				last_Teammates[i].y);
		//System.out.println("num teammates "+vis_Teammates);
		return(retval);
		}

	public int getPlayerNumber(long timestamp)
		{
		// use side effect of discovering number of visible teammates
		Vec2[] dummy = getTeammates(timestamp);
		// there are always exactly 5 robots on a team.
		return(getID()%(num_Teammates));
		}

	private	long	last_Obstaclest = 0;
	private	Vec2	last_Obstacles[];
	private	int	num_Obstacles;
	/**
	 * Get an array of Vec2s that point egocentrically from the
	 * center of the robot to the obstacles currently sensed by the 
	 * bumpers and sonars.
	 * @param timestamp only get new information 
	 *	if timestamp > than last call
	 *		or timestamp == -1 .
	 * @return the sensed obstacles.
	 */
	public Vec2[] getObstacles(long timestamp)
		{
		if((timestamp > last_Obstaclest)||(timestamp == -1))
			{
			if (timestamp != -1) last_Obstaclest = timestamp;
			Vec2 tmp_objs[] = new Vec2[all_objects.length];
			num_Obstacles = 0;
			/*--- check all objects ---*/
			for(int i = 0; i<all_objects.length; i++)
				{
				/*--- check if it's an obstacle and not self ---*/
				if (all_objects[i].isObstacle() &&
					(all_objects[i].getID() != unique_id))
					{
					Vec2 tmp = all_objects[i].getClosestPoint(
							position);
					if (tmp.r<obstacle_rangeM)
					tmp_objs[num_Obstacles++] = tmp;
					}
				}
			last_Obstacles = new Vec2[num_Obstacles];
			for(int i = 0; i<num_Obstacles; i++)
			last_Obstacles[i] = new Vec2(tmp_objs[i].x,
				tmp_objs[i].y);
			}
		Vec2[] retval = new Vec2[num_Obstacles];
		for(int i = 0; i<num_Obstacles; i++)
			retval[i] = new Vec2(last_Obstacles[i].x,
				last_Obstacles[i].y);
		return(retval);
		}


	private	double	kin_rangeM = 4.0;
	/**
	 * Set the maximum range at which a sensor reading should be considered
	 * kin.  Beyond this range, the readings are ignored.
	 * The default range on startup is 1 meter.
	 * @param range the range in meters.
	 */
	public void setKinMaxRange(double range)
		{
		kin_rangeM = range;
		}

	private	double	obstacle_rangeM = 4.0;
	/**
	 * Set the maximum range at which a sensor reading should be considered
	 * an obstacle.  Beyond this range, the readings are ignored.
	 * The default range on startup is 1 meter.
	 * @param range the range in meters.
	 */
	public void setObstacleMaxRange(double range)
		{
		obstacle_rangeM = range;
		}


	/**
	 * Get the position of the robot in global coordinates.
	 * @param timestamp only get new information 
	 *	if timestamp > than last call or timestamp == -1.
	 * @return the position.
	 * @see Simple#resetPosition
	 */
	public Vec2 getPosition(long timestamp)
		{
		return(new Vec2(position.x, position.y));
		}

		
	/**
	 * Get the position of the robot in global coordinates.
	 * @return the position.
	 */
	public Vec2 getPosition()
		{
		return(new Vec2(position.x, position.y));
		}
		

	/**
	 * Reset the odometry of the robot in global coordinates.
	 * This might be done when reliable sensor information provides
	 * a very good estimate of the robot's location, or if you
	 * are starting the robot in a known location other than (0,0).
	 * Do this only if you are certain you're right!
	 * @param position the new position.
	 * @see Simple#getPosition
	 */
	public void resetPosition(Vec2 posit)
		{
		position.setx(posit.x);
		position.setx(posit.y);
		}


	/**
	*/
	public double getSteerHeading(long timestamp)
		{
		return(steer.t);
		}

	
	/**
	*/
	public void resetSteerHeading(double heading)
		{
		/* ensure in legal range */
		heading = Units.ClipRad(heading); 

		/* set the angle */
		steer.sett(heading);
		}


	private double desired_heading;
	/**
	*/
	public void setSteerHeading(long timestamp, double heading)
		{
		/* ensure in legal range */
		desired_heading = Units.ClipRad(heading);
		}
	
	
	private double desired_speed = 0;
	/**
	@see SealedSphereSim#setSpeed
	*/ 
	public void setSpeed(long timestamp, double speed)
		{
		/* ensure legal range */
		if (speed > 1.0) speed = 1.0;
		else if (speed < 0) speed = 0;
		desired_speed = speed;
		}


	private double base_speed = MAX_TRANSLATION;
	/**
	@see SealedSphereSim#setBaseSpeed
	*/
	public void setBaseSpeed(double speed)
		{
		if (speed > MAX_TRANSLATION) speed = MAX_TRANSLATION;
		else if (speed < 0) speed = 0;
		base_speed = speed;
		}


        /**
        NOT IMPLEMENTED
        */
        public void receive(Message m)
                {
                /* todo */
                }
	}
