/*
 * JavaBotSim.java
 */

package JavaBotSim;

import java.io.*;
import java.awt.*;
import java.net.*;
import java.awt.event.*;
import EDU.gatech.cc.is.abstractrobot.*;
import EDU.gatech.cc.is.clay.*;
import EDU.gatech.cc.is.util.*;

/**
 * Application that runs a control system in
 * simulation.
 * <P>
 * To run this program, first ensure you are in the correct directory (where
 * the JavaBotSim.class file is), then type "java JavaBotSim".
 * <P>
 * For more detailed information, see the
 * <A HREF="../../../../../JavaBotSim/docs/index.html">JavaBotSim page</A>.
 * <P>
 * <A HREF="../COPYRIGHT.html">Copyright</A> 
 * (c)1997 Georgia Tech Research Corporation
 *
 * @author Tucker Balch
 * @version $Revision: 1.6 $
 */

public class JavaBotSim extends Frame 
	implements ActionListener, ItemListener
	{
	private static SimulationCanvas simulation;
	private static String	dsc_file;
	private	static	int	height = 500;
	private	static	int	width = 500;
	private	String	current_directory;
	//private	CheckboxMenuItem view_graphics;
	private	CheckboxMenuItem robot_ids;
	private	CheckboxMenuItem robot_trails;
	private	CheckboxMenuItem robot_state;

	/**
	 * Set up the frame and buttons.
	 */
	public JavaBotSim()
		{
		super("JavaBotSim");

		/*--- Set the title ---*/
		// try to find our hostnam first
		InetAddress this_host;
		try
			{
			this_host = InetAddress.getLocalHost();
			}
		catch (Exception e)
			{
			this_host = null;
			}
		String host_name = "unknown host";
		if (this_host != null)
			host_name = this_host.getHostName();
		setTitle("JavaBotSim"+" ("+host_name+")");

		/*--- Set up the menu bar ---*/
		MenuBar mb = new MenuBar();

		Menu mf = new Menu("File");
		mf.add("Load");
		mf.add("Print");
		mf.add("Quit");
		mf.addActionListener(this);
		mb.add(mf);

		Menu mv = new Menu("View");
		//view_graphics = new CheckboxMenuItem("Graphics");
		//mv.add(view_graphics);
		//view_graphics.addItemListener(this);
		robot_ids = new CheckboxMenuItem("Robot IDs");
		mv.add(robot_ids);
		robot_ids.addItemListener(this);
		robot_trails = new CheckboxMenuItem("Robot Trails");
		mv.add(robot_trails);
		robot_trails.addItemListener(this);
		robot_state = new CheckboxMenuItem("Robot State");
		mv.add(robot_state);
		robot_state.addItemListener(this);
		mb.add(mv);

		Menu hm = new Menu("Help");
		hm.add("Runtime Stats");
		hm.add("About");
		hm.add("     ");
		hm.addActionListener(this);

		mb.setHelpMenu(hm);
	
		setMenuBar(mb);

		/*--- Set up the buttons ---*/
		Panel button_area = new Panel();
		Button start_button = new Button("Reset/Reload");
		start_button.addActionListener(this);
		button_area.add(start_button);
		Button resume_button = new Button("Start/Resume");
		resume_button.addActionListener(this);
		button_area.add(resume_button);
		Button pause_button = new Button("Pause");
		pause_button.addActionListener(this);
		button_area.add(pause_button);
		add("North",button_area);

		/*--- Set up the Graphical Area ---*/
		Panel playing_field_panel = new Panel();
		simulation = 
			new SimulationCanvas((Frame)this,width,height,dsc_file);
		playing_field_panel.add(simulation);
		add("South",playing_field_panel);

		/*--- Pack Everything ---*/
		pack();
		setResizable(false);

		/*--- tell the simulation to load and run ---*/
		if (dsc_file!=null) simulation.reset();
		if (simulation.descriptionLoaded())// only if loaded ok.
			{
			simulation.start();
			}
		}

	/**
	 * Handle checkbox events
	 */
	public void itemStateChanged(ItemEvent e)
		{
		String item = e.getItem().toString();

		//if (item == "Graphics")
		//	{
		//	simulation.setGraphics(view_graphics.getState());
		//	}
		if (item == "Robot IDs")
			simulation.setDrawIDs(robot_ids.getState());
		if (item == "Robot Trails")
			simulation.setDrawTrails(robot_trails.getState());
		if (item == "Robot State")
			simulation.setDrawState(robot_state.getState());
		}

	/**
	 * Handle button pushes.
	 */
	public void actionPerformed(ActionEvent e)
		{
		String command = e.getActionCommand();

		/*--- Load ---*/
		if (command=="Load")
			{
			FileDialog fd = new FileDialog(this,
				"Load New Description File",
				FileDialog.LOAD);

			/*--- try to filter based on extension ---*/
			// curently, this has no effect under linux and Irix
			FilenameFilter filt = 
				new FilenameFilterByEnding(".dsc");
			fd.setFilenameFilter(filt);
			if (current_directory != null)
				fd.setDirectory(current_directory);
			fd.show();
			String tmpname = fd.getFile();
			if (tmpname==null) tmpname = "";
			current_directory = fd.getDirectory();
			if (current_directory!=null) 
				tmpname = current_directory.concat(tmpname);
			if (tmpname!=null)
				{
				dsc_file = tmpname;
				simulation.load(dsc_file);
				}
			}

		/*--- Quit ---*/
		if (command=="Quit")
			{
			simulation.quit();
			System.exit(0);
			}

		/*--- Runtime Stats ---*/
		if (command=="Runtime Stats")
			{
			simulation.showRuntimeStats();
			}

		/*--- About ---*/
		else if(command == "About")
			{
			Dialog tmp = new DialogMessage((Frame)this,
				"About JavaBotSim",
				JavaBotsVersion.longReport() + "\n");
			}

		/*--- Avoid Cursor ---*/
		else if(command == "     ")
			{
			Dialog tmp = new DialogMessageJoke((Frame)this,
				"Avoid Cursor",
				"An implementation of the avoid_cursor\nmotor schema.");
			}

		/*--- Robot IDs ---*/
		else if(command == "Robot IDs")
			{
			System.out.println(robot_ids.getState());
			}

		/*--- Print ---*/
		else if (command == "Print")
			{
			PrintJob pjob = getToolkit().getPrintJob(this,
				"Print?", null);
			if (pjob != null)
				{
				Graphics pg = pjob.getGraphics();
				if (pg != null)
					printAll(pg);
					pg.dispose();
				}
			pjob.end();
			}

		/*--- Start/Resume ---*/
		else if (command == "Start/Resume")
			{
			simulation.start();
			}

		/*--- Reset ---*/
		else if (command == "Reset/Reload")
			{
			simulation.reset();
			}

		/*--- Pause ---*/
		else if (command == "Pause")
			{
			simulation.pause();
			}
		}

        /**
	 * Main for JavaBotSim.
         */
	public static void main(String[] args)
		{
		long runtime = 0;
		
		/*--- check the arguments ---*/
		if (args.length >= 1)
			{
			if (args[0].equalsIgnoreCase("-version"))
				{
				System.out.println(
					JavaBotsVersion.longReport());
				System.exit(0);
				}
			else
				dsc_file=args[0];
			}
		System.out.println(JavaBotsVersion.shortReport());
		try
			{
			if (args.length >= 2)
				width = Integer.parseInt(args[1]);

			if (args.length >= 3)
				height = Integer.parseInt(args[2]);
			}
		catch (Exception e)
			{
			System.out.println(
				"usage: JavaBotSim [-version] descriptionfile width height");
			}
		/*--- make the window ---*/
		JavaBotSim jbs = new JavaBotSim();
		jbs.show();
		}
	}
/*
$Log: JavaBotSim.java,v $
Revision 1.6  1998/02/24 00:21:45  tucker
*** empty log message ***

Revision 1.5  1998/02/23 22:57:31  tucker
fixed bugs

Revision 1.4  1998/02/04 17:48:57  tucker
*** empty log message ***

Revision 1.3  1998/02/02 05:54:05  tucker
*** empty log message ***

# Revision 1.2  1998/02/02  04:52:23  tucker
# fixed class naming bug
#
# Revision 1.1  1998/01/30  22:56:18  tucker
# Initial revision
#
Revision 1.11  1997/10/29 21:30:13  tucker
added version stuff.

Revision 1.10  1997/10/14 00:43:44  tucker
removed revision number from titlebar

Revision 1.9  1997/08/17 16:59:54  tucker
*** empty log message ***

# Revision 1.8  1997/08/17  16:33:30  tucker
# massaged comments
#
Revision 1.7  1997/08/07 15:12:04  tucker
modified copyright info

Revision 1.6  1997/08/05 23:45:03  tucker
*** empty log message ***

# Revision 1.5  1997/08/05  23:26:10  tucker
# *** empty log message ***
#
# Revision 1.4  1997/08/05  23:16:42  tucker
# *** empty log message ***
#
# Revision 1.3  1997/08/05  22:47:34  tucker
# added reference to version information
#
# Revision 1.2  1997/08/05  22:44:40  tucker
# nothing
#
# Revision 1.1  1997/08/05  22:41:26  tucker
# Initial revision
#
*/
