/*
 * SimulationCanvas.java
 */

package JavaBotSim;

import java.io.*;
import java.awt.*;
import java.lang.System;
import java.lang.Class;
import java.lang.reflect.*;
import EDU.gatech.cc.is.abstractrobot.*;
import EDU.gatech.cc.is.simulation.*;
import EDU.gatech.cc.is.util.*;

/**
 * Used within JavaBotSim to control and draw a simulation.
 * <P>
 * For more detailed information, see the
 * <A HREF="../../../../../JavaBotSim/docs/index.html">JavaBotSim page</A>.
 * <P>
 * <A HREF="../COPYRIGHT.html">Copyright</A>
 * (c)1997 Georgia Tech Research Corporation
 *
 * @author Tucker Balch
 * @version $Revision: 1.8 $
 */

public class SimulationCanvas extends Canvas implements Runnable
	{
	private Graphics g;
	private Frame parent;
	private	int height, width;
	private	Color bgcolor = new Color(0xFFFFFF);
	private	Image buffer;
	private	Graphics bufferg;
	private	boolean read_once = false; //indicates if we've read a dsc file
	private	boolean pause = true;
	private	boolean graphics_on = true;
	private SimulatedObject simulated_objects[] = new SimulatedObject[0];
	private ControlSystemS control_systems[] = new ControlSystemS[0];
	private	double top, bottom, left, right;
	private	double time_compression=1;
	private	long current_time = 0;
	private long sim_time = 0;
	private	long maxtimestep=100;
	private	long timeout=-1;
	private	long seed=-1;
	private	int trials=-1;
	private	Thread run_sim_thread;
	private String descriptionfile;
	private	int	idcounter = 0;
	private	boolean	to_draw = false;
	private	boolean	draw_ids = false; //don't draw robot ids
	private	boolean	draw_state = false; //don't draw robot state
	private	boolean	draw_trails = false; //don't draw object trails
	private long	startrun = 0;
	private long	frames = 0;

	/**
	 * The maximum number of objects in a simulation.
	 */
	public	static final int MAX_SIM_OBJS = 1000;

	/**
	 * Read the description of the world from a file.
	 */
	private	void loadEnvironmentDescription() throws IOException
		{
		FileReader file = new FileReader(descriptionfile);
		StreamTokenizer in = new StreamTokenizer(file);
		String token;
		SimulatedObject[] temp_objs = new SimulatedObject[MAX_SIM_OBJS];
		int temp_objs_count = 0;
		ControlSystemS[] temp_css = new ControlSystemS[MAX_SIM_OBJS];
		int temp_css_count = 0;
		double	x, y, t, r;
		int	color1, color2;
		int	vc;
		idcounter = 0;
		String	string1, string2;

		/*--- assume success. reset later if failure ---*/
		boolean dfl = true; //description_file_loaded;

		/*--- set default bounds before reading ---*/
		top = 5;
		bottom = -5;
		left = -5;
		right = 5;

		/*--- set up tokenizer ---*/
		in.wordChars('A','_'); // let _ be a word character
		in.quoteChar('"');     // " is the quote char

		/*--- tokenize the file ---*/
		token = "beginning of file";
		try
		{
		while (in.nextToken() != StreamTokenizer.TT_EOF)
			{
			if (in.ttype == StreamTokenizer.TT_WORD)
				{
				token = in.sval;
				if (false) System.out.println(token);
				
				/*--- it is a background statement ---*/
				//FORMAT: background color
				if (token.equals("background"))
					{
					if (in.nextToken()==
						StreamTokenizer.TT_WORD)
						{
						String tmp = in.sval;
						tmp = tmp.replace('x','0');
						tmp = tmp.replace('X','0');
						bgcolor = new Color(
							Integer.parseInt(tmp,16));
						}
					else
						{
						bgcolor = new Color((int)in.nval);
						}
					}

				/*--- it is a time statement ---*/
				//FORMAT: time accel_rate
				if (token.equals("time"))
					{
					if (in.nextToken()==
						StreamTokenizer.TT_NUMBER)
						{
						time_compression = in.nval;
						}
					else
						{
						token = in.sval; // for error report
						throw new IOException();
						}
					}

				/*--- it is a timeout statement ---*/
				//FORMAT: timeout time
				if (token.equals("timeout"))
					{
					if (in.nextToken()==
						StreamTokenizer.TT_NUMBER)
						{
						timeout = (long) in.nval;
						}
					else
						{
						token = in.sval; // for error report
						throw new IOException();
						}
					}

				/*--- it is a seed statement ---*/
				//FORMAT: seed seed_val
				if (token.equals("seed"))
					{
					if (in.nextToken()==
						StreamTokenizer.TT_NUMBER)
						{
						// skip for subsequent trials
						if (!read_once)
							seed = (long) in.nval;
						}
					else
						{
						token = in.sval; // for error report
						throw new IOException();
						}
					}

				/*--- it is a graphics statement ---*/
				//FORMAT: graphics on/off
				if (token.equals("graphics"))
					{
					if (in.nextToken()==
						StreamTokenizer.TT_NUMBER)
						{
						token = in.sval; // for error report
						throw new IOException();
						}
					else
						{
						if (in.sval.equals("off"))
							graphics_on = false;
						}
					}

				/*--- it is a trials statement ---*/
				//FORMAT: trials num_trials
				if (token.equals("trials"))
					{
					if (in.nextToken()==
						StreamTokenizer.TT_NUMBER)
						{
						if (trials == -1)
							trials = (int) in.nval;
						if (trials<0)
							throw new IOException();
						}
					else
						{
						token = in.sval; // for error report
						throw new IOException();
						}
					}

				/*--- it is a maxtimestep statement ---*/
				//FORMAT: maxtimestep milliseconds
				if (token.equals("maxtimestep"))
					{
					if (in.nextToken()==
						StreamTokenizer.TT_NUMBER)
						{
						maxtimestep = (long) in.nval;
						}
					else
						{
						token = in.sval; // for error report
						throw new IOException();
						}
					}

				/*--- it is a bounds statement ---*/
				//FORMAT: bounds left right bottom top
				if (token.equals("bounds"))
					{
					if (in.nextToken()==StreamTokenizer.TT_NUMBER)
						left = in.nval;
					else
						{
						token = in.sval; // for error report
						throw new IOException();
						}
					if (in.nextToken()==StreamTokenizer.TT_NUMBER)
						right = in.nval;
					else
						{
						token = in.sval; // for error report
						throw new IOException();
						}
					if (in.nextToken()==StreamTokenizer.TT_NUMBER)
						bottom = in.nval;
					else
						{
						token = in.sval; // for error report
						throw new IOException();
						}
					if (in.nextToken()==StreamTokenizer.TT_NUMBER)
						top = in.nval;
					else
						{
						token = in.sval; // for error report
						throw new IOException();
						}
					}

				/*--- it is an object statement ---*/
				//FORMAT object objectclass 
				//	x y t r color1 color2 visionclass
				if (token.equals("object"))
					{
					if (in.nextToken()==StreamTokenizer.TT_WORD)
						string1 = in.sval;
					else
						{
						token = in.sval; // for error report
						throw new IOException();
						}
					if (in.nextToken()==StreamTokenizer.TT_NUMBER)
						x = in.nval;
					else
						{
						token = in.sval; // for error report
						throw new IOException();
						}
					if (in.nextToken()==StreamTokenizer.TT_NUMBER)
						y = in.nval;
					else
						{
						token = in.sval; // for error report
						throw new IOException();
						}
					if (in.nextToken()==StreamTokenizer.TT_NUMBER)
						t = in.nval;
					else
						{
						token = in.sval; // for error report
						throw new IOException();
						}
					if (in.nextToken()==StreamTokenizer.TT_NUMBER)
						r = in.nval;
					else
						{
						token = in.sval; // for error report
						throw new IOException();
						}
					if (in.nextToken()==
						StreamTokenizer.TT_WORD)
						{
						String tmp = in.sval;
						tmp = tmp.replace('x','0');
						tmp = tmp.replace('X','0');
						color1 = Integer.parseInt(tmp,16);
						}
					else
						{
						color1 = (int)in.nval;
						}
					if (in.nextToken()==
						StreamTokenizer.TT_WORD)
						{
						String tmp = in.sval;
						tmp = tmp.replace('x','0');
						tmp = tmp.replace('X','0');
						color2 = Integer.parseInt(tmp,16);
						}
					else
						{
						color2 = (int)in.nval;
						}

					if (in.nextToken()==StreamTokenizer.TT_NUMBER)
						vc = (int)in.nval;
					else
						{
						token = in.sval; // for error report
						throw new IOException();
						}
					/*--- instantiate the obj ---*/
					token = string1; // in case of error
                        		Class rclass = Class.forName(string1);
                        		SimulatedObject
						obj = (SimulatedObject)rclass.newInstance();
					obj.init(x, y, t, r, new Color(color1),
						new Color(color2),vc,
						idcounter++,seed++);
					temp_objs[temp_objs_count++] = obj;
					}


				/*--- it is a robot statement ---*/
				//FORMAT robot robotclass controlsystemclass 
				//	x y t color1 color2 visionclass
				if (token.equals("robot"))
					{
					if (in.nextToken()==StreamTokenizer.TT_WORD)
						string1 = in.sval;
					else
						{
						token = in.sval; // for error report
						throw new IOException();
						}
					if (in.nextToken()==StreamTokenizer.TT_WORD)
						string2 = in.sval;
					else
						{
						token = in.sval; // for error report
						throw new IOException();
						}
					if (in.nextToken()==StreamTokenizer.TT_NUMBER)
						x = in.nval;
					else
						{
						token = in.sval; // for error report
						throw new IOException();
						}
					if (in.nextToken()==StreamTokenizer.TT_NUMBER)
						y = in.nval;
					else
						{
						token = in.sval; // for error report
						throw new IOException();
						}
					if (in.nextToken()==StreamTokenizer.TT_NUMBER)
						t = in.nval;
					else
						{
						token = in.sval; // for error report
						throw new IOException();
						}
					if (in.nextToken()==
						StreamTokenizer.TT_WORD)
						{
						String tmp = in.sval;
						tmp = tmp.replace('x','0');
						tmp = tmp.replace('X','0');
						color1 = Integer.parseInt(tmp,16);
						}
					else
						{
						color1 = (int)in.nval;
						}
					if (in.nextToken()==
						StreamTokenizer.TT_WORD)
						{
						String tmp = in.sval;
						tmp = tmp.replace('x','0');
						tmp = tmp.replace('X','0');
						color2 = Integer.parseInt(tmp,16);
						}
					else
						{
						color2 = (int)in.nval;
						}

					if (in.nextToken()==StreamTokenizer.TT_NUMBER)
						vc = (int)in.nval;
					else
						{
						token = in.sval; // for error report
						throw new IOException();
						}

					/*--- the robot ---*/
					token = string1; // in case of error
                        		Class rclass = Class.forName(string1);
                        		SimulatedObject
						obj = (SimulatedObject)rclass.newInstance();
					obj.init(x, y, t, 0, new Color(color1),
						new Color(color2),vc,
						idcounter++,seed++);
					temp_objs[temp_objs_count++] = obj;

					/*--- the control system ---*/
					token = string2; // in case of error
                        		Class csclass = Class.forName(string2);
					ControlSystemS css = 
						(ControlSystemS)
							csclass.newInstance();
					css.init((Simple)obj,seed++);
					//css.Configure();//save for later
					temp_css[temp_css_count++] = 
						(ControlSystemS) css;
					}
				}
			else
				{
				throw new IOException();
				}
			}

		/*--- catch any exceptions thrown in the parsing ---*/
		}
		catch (IOException e)
			{
			dfl = false;
			simulated_objects = new SimulatedObject[0];
			String msg =
				"bad format"+
				" at line "+ in.lineno() +
				" in " + descriptionfile +
				" near "+ 
				"'"+token+"'";
			Dialog tmp;
			if (graphics_on) tmp = new DialogMessage(parent,
				"JavaBotSim Error",msg);
			else
				System.out.println(msg);
			descriptionfile = null;
			}
                catch (ClassNotFoundException e )
                        {
			dfl = false;
			simulated_objects = new SimulatedObject[0];
			String msg =
				"unable to find class "+
				"'"+token+"'" +
				" at line "+ in.lineno() +
				" in " + descriptionfile +".\n"+
				"You may need to check your CLASSPATH\n"+
				"or start JavaBotSim in another directory.";
			Dialog tmp;
			if (graphics_on) tmp = new DialogMessage(parent,
				"JavaBotSim Error",msg);
			else
				System.out.println(msg);
			descriptionfile = null;
                        }
                catch (IllegalAccessException e)
                        {
			dfl = false;
			simulated_objects = new SimulatedObject[0];
			String msg = 
				"illegal to access class "+
				"'"+token+"'" +
				" at line "+ in.lineno() +
				" in " + descriptionfile;
			Dialog tmp;
			if (graphics_on) tmp = new DialogMessage(parent,
				"JavaBotSim Error", msg);
			else
				System.out.println(msg);
			descriptionfile = null;
                        }
                catch (InstantiationException e)
                        {
			dfl = false;
			simulated_objects = new SimulatedObject[0];
			String msg =
				"instantiation error for "+
				"'"+token+"'" +
				" at line "+ in.lineno() +
				" in " + descriptionfile ;
			Dialog tmp;
			if (graphics_on) tmp = new DialogMessage(parent,
				"JavaBotSim Error", msg);
			else
				System.out.println(msg);
			descriptionfile = null;
                        }
		catch (ClassCastException e)
			{
			dfl = false;
			simulated_objects = new SimulatedObject[0];
			String msg =
				"class conflict for "+
				"'"+token+"'" +
				" at line "+ in.lineno() +
				" in " + descriptionfile +"."+
				" It could be that the control system was not "+
				" written for the type of robot you " +
				" specified.";
			Dialog tmp;
			if (graphics_on) tmp = new DialogMessage(parent,
				"JavaBotSim Error",msg);
			else
				System.out.println(msg);
			descriptionfile = null;
			}

		/*--- set up global arrays of objs and cont systems ---*/
		simulated_objects = new SimulatedObject[temp_objs_count];
		for(int i=0; i<temp_objs_count; i++)
			simulated_objects[i] = temp_objs[i];
		for(int i=0; i<temp_objs_count; i++)
			{	
			// let everyone take a step to update their pointers
			simulated_objects[i].takeStep(0,simulated_objects);
			}
		control_systems = new ControlSystemS[temp_css_count];
		for(int i=0; i<temp_css_count; i++)
			{
			control_systems[i] = temp_css[i];
			control_systems[i].configure();
			}

		description_file_loaded = dfl;
		read_once = true;
		}


	/**
	Set up the SimulationCanvas.
	*/
	public SimulationCanvas(Frame p, int w, int h,
		String dscfile)
		{
		if (p == null) 
			{
			graphics_on = false;
			pause = false;
			}
		else
			{
			graphics_on = true;
			pause = true;
			}
		parent = p;
		simulated_objects = new SimulatedObject[1];
		control_systems = new ControlSystemS[1];
		width = w;
		height = h; 
		descriptionfile = dscfile;

		if (graphics_on)
			{
			setBounds(0,0,width,height);
			setBackground(Color.white);
			}

		/*--- instantitate thread ---*/
		run_sim_thread = new Thread(this);
		run_sim_thread.start();
		}


	private boolean description_file_loaded = false;
	/**
	 * Provide info about whether we have successufully
	 * loaded the file.
	 * @return true if a file is loaded, false otherwise.
	 */
	public boolean descriptionLoaded()
		{
		return (description_file_loaded);
		}


	private boolean keep_running = true;
	/**
	 * Run the simulation.
	 */
	public void run()
		{
		//pause = true;
		long start_time = System.currentTimeMillis();
		long last_time = System.currentTimeMillis();
		long sim_timestep = 0;
		boolean robots_done = false;
		while (keep_running)
			{
			while(pause||(description_file_loaded==false))
				{
				if (graphics_on) this.repaint();
				try {Thread.sleep(200);}
				catch(InterruptedException e){}
				last_time = System.currentTimeMillis()-1;
				}
			current_time = System.currentTimeMillis();
			sim_timestep = (long)(
				(double)(current_time - last_time)*
				time_compression);
			if (sim_timestep>maxtimestep)
				sim_timestep = maxtimestep;

			/*--- run control systems and check for done ---*/
			robots_done = true;
			for(int i=0; i<control_systems.length; i++)
				{
				int stat = control_systems[i].takeStep();
				if (stat!=ControlSystemS.CSSTAT_DONE)
					robots_done = false;
				}

			/*--- run the physics ---*/
			for(int i=0; i<simulated_objects.length; i++)
				simulated_objects[i].takeStep(
					sim_timestep, 
					simulated_objects);
			last_time = current_time;

			/*--- draw everything ---*/
			to_draw = true;
			if (graphics_on) this.repaint();
			if (to_draw&&graphics_on)
				{
				try {Thread.sleep(10);}
				catch(InterruptedException e){}
				}

			/*--- garbage collect every time ---*/
			// this is to make cycle times more homogeneous
			//System.gc();  // too slow!

			/*--- count frames ---*/
			frames++;// for statistics gathering

			/*--- check for timeout or done ---*/
			if (((timeout>0)&&(sim_time>=timeout))
				|| robots_done)
				{
				if (trials <= 1)
					{
					for (int i = 0; 
						i<control_systems.length; i++)
						{
						control_systems[i].trialEnd();
						control_systems[i].quit();
						}
					keep_running=false;
					if (graphics_on == false)
						showRuntimeStats();
					System.exit(0);
					}
				else
					{
					for (int i = 0; 
						i<control_systems.length; i++)
						{
						control_systems[i].trialEnd();
						}
					trials--;
					sim_time = 0;
					reset();
					start();
					}
				}

			/*--- increment simulation time ---*/
			sim_time += sim_timestep;
			}
		}


	/**
	 * Handle a drawing request.
	 */
	public synchronized void update(Graphics g)
		{
		if((bufferg!=null)&&(graphics_on))
			{
			bufferg.setColor(bgcolor);
			bufferg.fillRect(0,0,width,height);

			/*--- draw trails first ---*/
			for(int i= 0; i<simulated_objects.length; i++)
				{
				// if robot
				if (simulated_objects[i] instanceof Simple)
					{
					// draw trail
					if (draw_trails)
						{
					simulated_objects[i].drawTrail(bufferg,
						width,height,
						top, bottom,left,right);
						}
					}
				}

			/*--- draw IDs and state ---*/
			for(int i= 0; i<simulated_objects.length; i++)
				{
				// if robot
				if (simulated_objects[i] instanceof Simple)
					{
					if (draw_ids)
						{
					simulated_objects[i].drawID(bufferg,
						width,height,
						top, bottom,left,right);
						}
					if (draw_state)
						{
					simulated_objects[i].drawState(bufferg,
						width,height,
						top, bottom,left,right);
						}
					}

				/*--- draw the robot ---*/
				simulated_objects[i].draw(bufferg,width,height,
					top, bottom,left,right);
				}
			g.drawImage(buffer, 0, 0, this);
			}
		to_draw = false;
		}

		
	/**
	Resize the SimulationCanvas.
	*/
	public void setSize(int h, int w)
		{
		Dimension d = getPreferredSize();
		height = d.height;
		width = d.width;
		setBounds(0,0,width,height);
		}

	/**
	 * Handle a quit event.
	 */
	public void quit()
		{
		//call all the control system .quit methods
		for (int i = 0; i<control_systems.length; i++)
			{
			control_systems[i].trialEnd();
			control_systems[i].quit();
			}
		}

	/**
	 * Handle a reset event.
	 */
	public void reset()
		{
		if (graphics_on == true) 
			pause = true;
		else
			pause = false;

		if (descriptionfile!=null)
			{
			try
				{
				loadEnvironmentDescription();
				}
			catch (FileNotFoundException e)
				{
				Dialog tmp;
				description_file_loaded = false;
				simulated_objects = new SimulatedObject[0];
				String msg = "file not found: "
					+ descriptionfile;
				if (graphics_on)
					tmp = new DialogMessage(parent,
						"JavaBotSim Error", msg);
				else
					System.out.println(msg);
				descriptionfile = null;
				}
			catch(IOException e)
				{
				Dialog tmp;
				description_file_loaded = false;
				simulated_objects = new SimulatedObject[0];
				String msg = "error trying to load "+
					descriptionfile;
				if (graphics_on)
					tmp = new DialogMessage(parent,
					"JavaBotSim Error", msg);
				else
					System.out.println(msg);
				descriptionfile = null;
				}
			if (graphics_on)
				{
				buffer = createImage(width,height);
				bufferg = buffer.getGraphics();
				bufferg.setColor(Color.white);
				bufferg.fillRect(0,0,width,height);
				this.repaint();
				pause = true;
				}
			}
		else
			{
			Dialog tmp;
			String msg = "Error: no description file";
			if (graphics_on)
				tmp = new DialogMessage(parent,
				"JavaBotSim Error",
				"You must choose description file first.\n"+
				"Use the `load' option under the `file' menu.");
			}
		}


	/**
	 * Handle a start/resume event.
	 */
	public void start()
		{
		if (description_file_loaded)
			{
			pause = false;
			if (graphics_on) this.repaint();
			// tell the control systems the trial is beginning
			for (int i = 0; i<control_systems.length; i++)
				{
				control_systems[i].trialInit();
				}
			startrun = System.currentTimeMillis();
			frames = 0;
			}
		else
			{
			Dialog tmp;
			if (graphics_on) 
				tmp = new DialogMessage(parent,
				"JavaBotSim Error",
				"You must load a description file first.\n"+
				"Use the `load' option under the `file' menu.");
			}
		}


	/**
	 * Handle a Runtime Stats event
	 */
	public void showRuntimeStats()
		{
		long	f = frames;
		long	t = System.currentTimeMillis() - startrun;

		Runtime r = Runtime.getRuntime();

		String this_sim = 
			  " trial number	: "+trials+" (counts down)\n"+
			  " sim time    	: "+sim_time+" milliseconds\n"+
			  " maxtimestep 	: "+maxtimestep+" milliseconds\n"+
			  " timeout      	: "+timeout+" milliseconds\n";

		if (pause)
			{
			this_sim = this_sim +
			  " frames/second	: N/A while paused\n"
			+ " free memory  	: "+r.freeMemory()+"\n"
			+ " total memory	: "+r.totalMemory()+"\n"
			+ " os.name      	: "+System.getProperty("os.name")+"\n"
			+ " os.version   	: "+System.getProperty("os.version")+"\n"
			+ " os.arch      	: "+System.getProperty("os.arch")+"\n"
			+ " java.version	: "+System.getProperty("java.version")+"\n";
			}
		else
			{
			double rate = 1000*(double)frames/(double)t;
			this_sim = this_sim +
			  " frames/second	: "+rate+ "\n"
			+ " free memory	: "+r.freeMemory()+"\n"
			+ " total memory	: "+r.totalMemory()+"\n"
			+ " os.name	: "+System.getProperty("os.name")+"\n"
			+ " os.version	: "+System.getProperty("os.version")+"\n"
			+ " os.arch	: "+System.getProperty("os.arch")+"\n"
			+ " java.version	: "+System.getProperty("java.version")+"\n";
			}
		Dialog tmp;
		if (graphics_on)
			tmp = new DialogMessage(parent, "Runtime Stats",
				this_sim);
		else
			System.out.println(this_sim);
		}


	/**
	Handle a pause event.
	*/
	public void pause()
		{
		pause = true;
		}


	/**
	 * Handle setDrawIDs
	 */
	public void setDrawIDs(boolean v)
		{
		draw_ids = v;
		}


	/**
	 * Handle setGraphics
	 */
	public void setGraphics(boolean v)
		{
		graphics_on = v;
		}


	/**
	 * Handle setDrawState
	 */
	public void setDrawState(boolean v)
		{
		draw_state = v;
		}


	/**
	 * Handle setDrawTrails
	 */
	public void setDrawTrails(boolean v)
		{
		draw_trails = v;
		}


	/**
	Handle a load request.
	*/
	public void load(String df)
		{
		pause();
		descriptionfile = df;
		reset();
		}

	}
/*
$Log: SimulationCanvas.java,v $
Revision 1.8  1998/07/25 22:42:24  tucker
*** empty log message ***

# Revision 1.7  1998/06/02  20:52:41  tucker
# *** empty log message ***
#
Revision 1.6  1998/04/06 21:13:32  tucker
*** empty log message ***

Revision 1.5  1998/02/23 22:57:31  tucker
faster

Revision 1.4  1998/02/04 17:48:57  tucker
*** empty log message ***

Revision 1.3  1998/02/02 04:52:23  tucker
fixed class naming bug

# Revision 1.2  1998/01/30  22:56:18  tucker
# *** empty log message ***
#
Revision 1.1  1998/01/29 02:49:03  tucker
Initial revision

# Revision 1.6  1997/10/28  01:19:27  tucker
# added maxtimestep
#
Revision 1.5  1997/10/14 00:43:44  tucker
added capability to simulate Pebbles robot.

Revision 1.4  1997/09/09 19:14:50  tucker
*** empty log message ***

Revision 1.3  1997/08/17 17:00:06  tucker
*** empty log message ***

# Revision 1.2  1997/08/07  15:13:27  tucker
# *** empty log message ***
#
Revision 1.1  1997/08/05 22:54:55  tucker
Initial revision

*/
