/*
 * play.java
 */

import	EDU.gatech.cc.is.util.Vec2;
import	EDU.gatech.cc.is.abstractrobot.*;
import	EDU.gatech.cc.is.clay.*;

/**
 * <B>Introduction</B><BR>
 * Example of complex schema-based control system for a Pebbles robot.
 * It uses a motor schema-based configuration to search for a red
 * ball then kick it.
 * The configuration is built using Clay.
 * It can be run in simulation or on a robot.
 * <P>
 * For detailed information on how to configure behaviors, see the
 * <A HREF="../clay/docs/index.html">Clay page</A>.
 * <P>
 * <A HREF="../COPYRIGHT.html">Copyright</A>
 * (c)1997 Georgia Tech Research Corporation
 *
 * @author Tucker Balch
 * @version $Revision: 1.2 $
 */

public class play  extends ControlSystemPebbles
	{
	public final static boolean DEBUG = false;
	private NodeVec2	steering_configuration;
	i_FSA_ba STATE_MACHINE;

	/**
	Configure the control system using Clay.
	*/
	public void Configure()
		{

		//======
		// Set some initial hardware configurations.
		//======
		if (DEBUG) System.out.println("play .Configure()");
		abstract_robot.setObstacleMaxRange(3.0); // don't consider 
							// things further away
		abstract_robot.setBaseSpeed(1.0*abstract_robot.MAX_TRANSLATION);

		//======
		// perceptual schemas
		//======
		//--- robot's global position
		NodeVec2
		PS_GLOBAL_POS = new v_GlobalPosition_r(abstract_robot);

		//--- obstacles
		NodeVec2Array // the sonar readings
		PS_OBS = new va_Obstacles_r(abstract_robot);

		//--- homebase 
		// Don't use this now but it may be handy soon.
		NodeVec2      // a fixed position in front of the door
		PS_HOMEBASE_GLOBAL = new v_FixedPoint_(0.0, 0.0);
		NodeVec2      // make it egocentric
		PS_HOMEBASE = new v_GlobalToEgo_rv(abstract_robot,
				PS_HOMEBASE_GLOBAL);

		//--- heading
		NodeVec2
		PS_HEADING = new v_SteerHeading_r(abstract_robot);

		//--- balls
		NodeVec2Array 
		PS_RED_BALLS = 
			new va_VisualObjects_r(0,abstract_robot);
		NodeVec2
		PS_CLOSEST_RED_BALL = new v_Closest_va(PS_RED_BALLS);

		NodeVec2Array 
		PS_GREEN_BALLS = 
			new va_VisualObjects_r(1,abstract_robot);
		NodeVec2
		PS_CLOSEST_GREEN_BALL = new v_Closest_va(PS_GREEN_BALLS);

		
		//======
		// Perceptual Features
		//======
		// is red ball visible?
		NodeBoolean
		PF_RED_BALL_VISIBLE = new b_NonZero_v(PS_CLOSEST_RED_BALL);
		NodeBoolean
		PF_NOT_RED_BALL_VISIBLE = new b_Not_s(PF_RED_BALL_VISIBLE);

		//is a green ball visible?
		NodeBoolean
		PF_GREEN_BALL_VISIBLE = new b_NonZero_v(PS_CLOSEST_GREEN_BALL);
		NodeBoolean
		PF_NOT_GREEN_BALL_VISIBLE = new b_Not_s(PF_GREEN_BALL_VISIBLE);
		NodeBoolean  // remember green ball for 5 seconds
		PF_SCARED = new b_Persist_s(5.0, PF_GREEN_BALL_VISIBLE);
		NodeBoolean
		PF_NOT_SCARED = new b_Not_s(PF_SCARED);

		// close to red ball
		NodeBoolean
		PF_CLOSE_TO_RED_BALL = new b_Close_vv(
				abstract_robot.RADIUS+0.2, 
				PS_CLOSEST_RED_BALL,
				new v_FixedPoint_(0,0));
		NodeBoolean
		PF_NOT_CLOSE_TO_RED_BALL = new b_Not_s(PF_CLOSE_TO_RED_BALL);



		//======
		// motor schemas
		//======
		// go to the ball
		NodeVec2
		MS_GO_TO_RED_BALL = new v_LinearAttraction_v(0.0, 0.0,
			PS_CLOSEST_RED_BALL);

		// avoid obstacles
		NodeVec2
		MS_AVOID_OBSTACLES = new v_Avoid_va(1.5,
			abstract_robot.RADIUS + 0.1,
			PS_OBS);

		// swirl obstacles wrt red ball
		NodeVec2
		MS_SWIRL_OBSTACLES_RED_BALL = new v_Swirl_vav(2.0,
			abstract_robot.RADIUS + 0.22,
			PS_OBS,
			PS_CLOSEST_RED_BALL);

		// search vector
		NodeVec2
		MS_SEARCH_VECTOR = new v_Noise_(5);

		// swirl obstacles
		NodeVec2
		MS_SWIRL_OBSTACLES_SEARCH = new v_Swirl_vav(2.0,
			abstract_robot.RADIUS + 0.1,
			PS_OBS,
			MS_SEARCH_VECTOR);


		//======
		// AS_FREEZE
		//======
		v_StaticWeightedSum_va 
		AS_FREEZE = new v_StaticWeightedSum_va();

		AS_FREEZE.weights[0]  = 0.001;
		AS_FREEZE.embedded[0] = PS_HEADING;


		//======
		// AS_SEARCH
		//======
		v_StaticWeightedSum_va 
		AS_SEARCH = new v_StaticWeightedSum_va();

		AS_SEARCH.weights[0]  = 0.3;
		AS_SEARCH.embedded[0] = MS_AVOID_OBSTACLES;

		AS_SEARCH.weights[1]  = 0.3;
		AS_SEARCH.embedded[1] = MS_SEARCH_VECTOR;

		AS_SEARCH.weights[2]  = 0.4;
		AS_SEARCH.embedded[2] = MS_SWIRL_OBSTACLES_SEARCH;



		//======
		// AS_GO_TO_RED_BALL
		//======
		v_StaticWeightedSum_va 
		AS_GO_TO_RED_BALL = new v_StaticWeightedSum_va();

		AS_GO_TO_RED_BALL.weights[0]  = 0.7;
		AS_GO_TO_RED_BALL.embedded[0] = MS_AVOID_OBSTACLES;

		AS_GO_TO_RED_BALL.weights[1]  = 0.7;
		AS_GO_TO_RED_BALL.embedded[1] = MS_GO_TO_RED_BALL;

		AS_GO_TO_RED_BALL.weights[2]  = 0.7;
		AS_GO_TO_RED_BALL.embedded[2] = MS_SWIRL_OBSTACLES_RED_BALL;



		//======
		// AS_KICK_RED_BALL
		//======
		v_StaticWeightedSum_va 
		AS_KICK_RED_BALL = new v_StaticWeightedSum_va();

		AS_KICK_RED_BALL.weights[0]  = 1.0;
		AS_KICK_RED_BALL.embedded[0] = MS_AVOID_OBSTACLES;

		AS_KICK_RED_BALL.weights[1]  = 1.0;
		AS_KICK_RED_BALL.embedded[1] = MS_GO_TO_RED_BALL;

		AS_KICK_RED_BALL.weights[2]  = 1.0;
		AS_KICK_RED_BALL.embedded[2] = MS_SWIRL_OBSTACLES_RED_BALL;


		//======
		// STATE_MACHINE
		//======
		STATE_MACHINE = new i_FSA_ba();

		STATE_MACHINE.state = 0;
		
		// STATE 0 EXPLORE
		STATE_MACHINE.triggers[0][0]  = PF_SCARED;
		STATE_MACHINE.follow_on[0][0] = 1; // transition to FREEZE
		STATE_MACHINE.triggers[0][1]  = PF_RED_BALL_VISIBLE;
		STATE_MACHINE.follow_on[0][1] = 2; // transition to CHASE

		// STATE 1 FREEZE (scared)
		STATE_MACHINE.triggers[1][0]  = PF_NOT_SCARED;
		STATE_MACHINE.follow_on[1][0] = 0; // transition to EXPLORE

		// STATE 2 CHASE
		STATE_MACHINE.triggers[2][0]  = PF_SCARED;
		STATE_MACHINE.follow_on[2][0] = 1; // transition to FREEZE
		STATE_MACHINE.triggers[2][1]  = PF_CLOSE_TO_RED_BALL;
		STATE_MACHINE.follow_on[2][1] = 3; // transition to KICK
		STATE_MACHINE.triggers[2][2]  = PF_NOT_RED_BALL_VISIBLE;
		STATE_MACHINE.follow_on[2][2] = 0; // transition to EXPLORE

		// STATE 3 KICK
		STATE_MACHINE.triggers[3][0]  = PF_SCARED;
		STATE_MACHINE.follow_on[3][0] = 1; // FREEZE
		STATE_MACHINE.triggers[3][1]  = PF_NOT_CLOSE_TO_RED_BALL;
		STATE_MACHINE.follow_on[3][1] = 2; // CHASE
		STATE_MACHINE.triggers[3][2]  = PF_NOT_RED_BALL_VISIBLE;
		STATE_MACHINE.follow_on[3][2] = 0; // transition to EXPLORE


		//======
		// STEERING
		//======
		v_Select_vai
		STEERING = new v_Select_vai((NodeInt)STATE_MACHINE);

		STEERING.embedded[0] = AS_SEARCH;
		STEERING.embedded[1] = AS_FREEZE;
		STEERING.embedded[2] = AS_GO_TO_RED_BALL;
		STEERING.embedded[3] = AS_KICK_RED_BALL;

		steering_configuration = STEERING;
		}
		
	/**
	Called every timestep to allow the control system to
	run.
	*/
	public int TakeStep()
		{
		Vec2	result;
		double	dresult;
		long	curr_time = abstract_robot.getTime();
		Vec2	p;

		// STEER
		result = steering_configuration.Value(curr_time);
		abstract_robot.setSteerHeading(curr_time, result.t);
		abstract_robot.setSpeed(curr_time, result.r);

		// State display
		int state = STATE_MACHINE.Value(curr_time);
		if (state==0)
			abstract_robot.setDisplayString(
				"searching for a red ball");
		else if (state==1)
			abstract_robot.setDisplayString(
				"scared of the green thing");
		else if (state==2)
			abstract_robot.setDisplayString(
				"chasing the red ball!");
		else if (state==3)
			abstract_robot.setDisplayString(
				"kicking the red ball!");

		return(CSSTAT_OK);
		}
	
	}
